/* --------------------------------------------------------------------
   Module: users.c
   Subject: UserList v1.25
   Author: Audin Malmin
   --------------------------------------------------------------------
   Description: scans Citadel msg.dat file and creates a user listing
                 suitable for posting into a room.
                NOTE: #define NDEBUG to remove debugging code.
   --------------------------------------------------------------------
   History:    (insert new entries at top of list)
   YYmmmDD/nam description

   95dec12/GS  version number changed to 1.25
   95dec12/GS  #output variable added to users.cfg - see users.cfg
                for details
   95dec12/GS  CONFIGURATION FILE! See users.cfg for details, and the
                config(), pfopen() functions.
   95dec12/GS  converted the excludenames[] array to a linked list, to
                facilitate a configuration file
   95dec12/GS  added some more autoposter exclusions
   95dec12/GS  apparently I missed a big bug in main() that made the
                autoposter-detecter throw out ALL users. fixed.
   95dec11/GS  attempt at optimizing readMessage() - no significant
                speedup on my messagebase
   95dec11/GS  optimized GetStr() - it turned 15 seconds into 12 seconds!
                (call to feof() was taken out and replaced with comparison
                 to constant EOF, plus some unnecessary comparisons were
                 removed)
   95dec11/GS  removed unnecessary PutStr()
   95dec11/GS  another blind attempt at optimization -- using hashes
                to prevent calling stricmp so often. Didn't make a
                noticeable difference, but it's there in case it does
                make a difference on larger msgbases. It is rare that
                a user will have the same hash as another.
   95dec11/GS  moved the autoposter-detection logic to the output loop
                instead of the message-reading loop, hoping for a speed
                increase. No noticeable increase on my 2 meg messagebase.
                Seems like a more logical place to put it though, to
                prevent unnecessary repetition -- may speed up runs on
                larger bases.
   95dec11/GS  double column, flush left output
   95dec11/GS  floating point operations in the output were converted to
                long operations. This prevents the floating point code from
                being linked in, providing a substantially smaller executable.
                Numbers are almost the same, with a .01 maximum margin of
                error. I don't think anybody'll notice!
   95dec11/GS  comparison with board name in main() was redundant
   95dec11/GS  converted the usertable array code to linked list code,
                to get rid of the realloc() stuff. Provided only a
                small speed increase, but linked lists are cool! A merge
                sort is used to sort the list.
   95dec11/GS  in WriteMessage(), the autoposter-detecting code was
                unnecessarily being executed repeatedly. It was moved,
                and the performance increase was very significant --
                more significant with each person you add to the excludenames[]
                array.
   95dec11/GS  table-ized the autoposter-checking stuff. To add or remove,
                or just look at the names excluded, look for the excludenames[]
                array.
   95dec11/GS  ANSI-compliant main() function (int main(void)) ;)
   95dec11/GS  quickly ran code through a beautifier to make it easier to work
                with and look at ;) as well as some other cosmetic changes

   95dec07/MM  upgrade to version 1.20 from 1.141a
   95dec07/MM  added exclusion of AutoPoster('s) names
   95dec07/MM  attempted at "Press any key" prompt, failed
   95dec07/MM  file handling/output looks better
   95dec07/MM  "Username" changed to "User Name"
   95nov13/MM  USERS.LST handling changed slightly to look better in public
               message format
   95nov05/MM  string added to beginning of program to let user know what
               version he/she is using
   95nov05/MM  enhanced file handling/output for USERS.LST
   95nov03/MM  directory changes [from c:\cit\messages\msg.dat to
               d:\citadel\msg.dat
   95nov03/MM  #include "pragmas.h" has been removed; this is unneeded
   95nov03/MM  commentary has been altered slightly
   95nov03/MM  error message output changed slightly
   95nov03/MM  by-author-exclusion changed from "Skaro" to "Higher Ground"
   95nov03/MM  output file handling changed slightly [header for USERS.LST]
   ???????/MM  header removed from top of output file that used to state
               version number
   --------------------------------------------------------------------
   Wish list:
   nam description

   MM  spaz-up the output file handling [USERS.LST]
   -------------------------------------------------------------------- */


/* Audin's original commentary:

   userlist....

     you'll probably have to change this a little...
     at least the directories will be wrong... (look down a page)

   And don't laugh at my coding...  :)

   This code started out as my mailing list program... and this is it's
   fourth generation, i think. [Mailing list, Keyserve, nodelist, userlist]

   Thats why there are a few unused variables around... you can ignore um...
   or delete um if you feel like it...
*/


/* --- Includes ------------------------------------------------------- */

#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <dos.h>
#include  <conio.h>
#include  <time.h>
#include  <ctype.h>
#include  <dir.h>

/* --- Defines -------------------------------------------------------- */

/* --- Types ---------------------------------------------------------- */

typedef struct user {
    char username[31];
    int hash;
    long messages;
    long msgk;
    struct user* next;
} user;

typedef struct xuser {
    char name[31];
    struct xuser* next;
} xuser;

/* --- Constants ------------------------------------------------------ */

/* --- Prototypes ----------------------------------------------------- */

void startup(void);
unsigned long readMessage(void);
void WriteMessage(void);
int  GetStr(FILE *fil, char *str, int mlen);
int  sort(user* this, user* that);
void stripansi(char *str);
user* lmerge(user* p, user* q);
user* lsort (user* p);
int hash(const char* s);
void config(void);
FILE *pfopen(const char* name, const char* mode, const char* dirs);

/* --- Global Variables ----------------------------------------------- */

/* --- Local Variables ------------------------------------------------ */

/* config variables, with their defaults */
char inputmail [FILENAME_MAX] = "c:\\cit\\msg.dat";
char outputmail[FILENAME_MAX] = "users.lst";
int listtype = 0;
xuser* firstxuser = NULL;
xuser* currxuser = NULL;
int output = 2;

FILE   *mailin, *mailout;
time_t starttime, endtime;

char   messagetext[8193];
char   author[31];
char   node[31];
char   sig[81];
long   nmessages = 0;
long   messagek = 0;
int    numuser = 0;

user* firstuser = NULL;
user* curruser = NULL;

char* exename;

/* --- Implementation ------------------------------------------------- */

int main(int argc, char** argv) {
    int i;
    unsigned long good = 1;

    exename = argv[0];

    starttime = time(NULL);

    startup();

    while(good) {
        good = readMessage();
    }

    firstuser = lsort(firstuser);

    fprintf(mailout, "     User List                         msgs    msgb    AVb/msg    %%msgs   %%msgb\n");
    fprintf(mailout, "     3e0\n");

    for(i = 0, curruser = firstuser; curruser != NULL; curruser = curruser->next) {
        int criteria;

        #ifndef NDEBUG
            printf(".");
        #endif

        if (listtype > 0) {
            criteria = (((long)curruser->msgk*10000)/(long)messagek >= (long)12);
        } else if (listtype < 0) {
            criteria = ((-listtype) > i);
        } else {
            criteria = 1;
        }

        if (criteria) {
            long op1, op1r, op2, op2r, op3, op3r;
            int valid = 1;

            /* exclude unwanted `users' */
            for (currxuser = firstxuser; currxuser != NULL; currxuser = currxuser->next) {
                if (stricmp(curruser->username, currxuser->name) == 0)
                    valid = 0;
            }

            if (!valid) {
                #ifndef NDEBUG
                    printf("\nExcluding %s", curruser->username);
                #endif

                continue;
            }

            op1  = (((long)curruser->msgk*100)/((long)curruser->messages))/100;
            op1r = (((long)curruser->msgk*100)/((long)curruser->messages))%100;

            op2  = (((long)curruser->messages*10000/(long)nmessages))/100;
            op2r = (((long)curruser->messages*10000/(long)nmessages))%100;

            op3  = (((long)curruser->msgk*10000/(long)messagek))/100;
            op3r = (((long)curruser->msgk*10000/(long)messagek))%100;

            fprintf(mailout, "%3d. %-30.30s 3e0 %5ld 3e0 %6lu 3e0 %4lu.%02lu 3e0 %2lu.%02lu%% 3e0 %1lu.%02lu%%\n",
                ++i,
                curruser->username,
                curruser->messages,
                curruser->msgk,
                op1,
                op1r,
                op2,
                op2r,
                op3,
                op3r
            );
        }
    }

    fprintf(mailout, "\nList generated from %ld messages totaling %ld bytes.\n", nmessages, messagek);

    fclose(mailin);
    fclose(mailout);

    endtime = time(NULL);
    printf("\nTotal process took %lu seconds.", endtime - starttime);

    return 0;
}


/*
 * Read a message...
 */
unsigned long readMessage(void) {
    int c;

    author[0] = 0;
    sig[0]   = 0;

    /* find start of message */
    do {
		c = fgetc(mailin);
	} while (c != 0xff && c != EOF);

	if (c == EOF)
        return 0;

    for (;;) {
        int c2;

        c = fgetc(mailin);
        if (c == 'M' || c == EOF)
            break;

        switch (c) {
            case 'A':     messagek += GetStr(mailin, author,    30);  break;
            case '.':     messagek += GetStr(mailin, sig,       80);  break;
            case 'O':     messagek += GetStr(mailin, node,      30);  break;

            default:
                /* discard unknown field */
                c2 = fgetc(mailin);
                while (c2 && c2 != EOF)
                    c2 = fgetc(mailin);
                break;
        }
    };

    if (c == EOF)
        return 0;

    GetStr(mailin, messagetext, 8192);

    /* Only messages in netrooms get Node Signatures...  this is how we detect */
    /* local netmessages...  So you need to have at least one #sig setup in cit. */

    if(node[0] || sig[0]) {
        messagek += strlen(messagetext);
        nmessages++;
        WriteMessage();
    }

    return 1;
}


void WriteMessage(void) {
    int usernum, lpc;
    user* lp;
    user* newuser;
    static int column = 0;
    const int maxcol = 78;
    int h;

    stripansi(author);

#ifdef XYZZY
    /* exclude unwanted `users' */
    for (i = 0; excludenames[i] != NULL; ++i) {

        if (stricmp(author, excludenames[i]) == 0) {
            *author = 0;
            *messagetext = 0;
            return;
        }
    }
#endif

    h = hash(author);
    for (lpc = 0, lp = firstuser, usernum = -1; lpc < numuser; ++lpc, lp = lp->next) {
        if (h == lp->hash) {
            usernum = lpc;
            break;
        }
    }

    if (usernum==-1) {
        numuser++;
        usernum=numuser-1;

        if (curruser == NULL) {
            /* list has no elements so far */
            firstuser = curruser = (user *)malloc(sizeof (user));
            curruser->next = NULL;
        }
        else {
            /* list has at least one element */
            newuser = (user *)malloc(sizeof (user));
            newuser->next = NULL;
            curruser->next = newuser;
            curruser = newuser;
        }

        strcpy(curruser->username, author);
        curruser->hash = h;
        curruser->messages = 1;
        curruser->msgk = strlen(messagetext);

#ifdef XYZZY
        printf("Detected User: %25s\n", author);
        /*   printf("!");*/
#endif
        if (output == 1) {
            printf("\r%3d. %-30s", numuser, author);
        } else if (output == 2) {
            if (column + 33 > maxcol) {
                printf("\n");
                column = 0;
            }
            column += printf("%-30s   ", author);
        }
    }
    else {


        /* duh -- we don't need to traverse the list! the element we
                   need is already there! */
#ifdef XYZZY
        /* traverse the list to get to user number `usernum' */
        for (lpc = 0, newuser = firstuser; lpc < usernum; ++lpc, newuser = newuser->next)
            ;
#endif

        lp->messages++;
        lp->msgk += strlen(messagetext);
    }

    author[0] = 0;
    messagetext[0] = 0;
}


int sort(user* this, user* that) {
    if (this->msgk > that->msgk)
        return -1;
    else if (this->msgk < that->msgk)
        return 1;
    else return 0;
}


/* merge 2 lists under dummy head item */
user* lmerge (user* p, user* q) {
    user* r;
    user head;

    for (r = &head; p && q; ) {
        if ( sort(p, q) < 0 ) {
            r = r->next = p;
            p = p->next;
        }
        else {
            r = r->next = q;
            q = q->next;
        }
    }
    r->next = (p ? p : q);
    return head.next;
}


/* split list into 2 parts, sort each recursively, merge */
user* lsort (user* p) {
    user* q;
    user* r;

    if ( p ) {
        q = p;
        for ( r = q->next; r && (r = r->next) != NULL; r = r->next )
            q = q->next;
        r = q->next;
        q->next = NULL;
        if ( r )
            p = lmerge(lsort(r), lsort(p));
    }
    return p;
}


/* -------------------------------------------------------------------- */
/* GetStr()        gets a null-terminated string from a file           */
/* -------------------------------------------------------------------- */
int GetStr(FILE *fil, char *str, int mlen) {

    int   l = 0;
    int ch;

    ch = fgetc(fil);
    while (ch && ch != EOF && l < mlen) {
        /* if (ch != 0xFF && ch != '\r') */
            str[l++]=ch;
        ch = fgetc(fil);
    }

    str[l]='\0';

    return(l);
}


/* changes string being sent, does not return a pointer like deansi */
/* And this was taken from original GremCit code... */
void stripansi(char *str) {
    int i=0;

    char *pc;
    pc = str;

    while (*pc) {
        if (*pc == 1)   pc += 2;                  /* Zap Ctrl-A plus following Char */
        else {
            str[i] = *pc;
            i++;
            pc++;
        }
    }
    str[i] = 0;
}


/* -------------------------------------------------------------------- */
/* startup...  opens all files and initializes hashes, etc.             */
/* -------------------------------------------------------------------- */
void startup(void) {
    printf("USERS v1.25 by Magrathea, Mystery Man, and Golden Spud.\n"
           "Scans a message base for netusers and creates a sorted list.\n"
           "------------------------------------------------------------\n");

    config();

    if((mailin = fopen(inputmail, "rb")) == NULL) {
        printf("\n\nFATAL: Cannot open msg.dat file\n");
        exit(1);
    }

    setvbuf(mailin, NULL, _IOFBF, 4096);

    if((mailout = fopen(outputmail, "wt")) == NULL) {
        printf("\n\nFATAL: Cannot open output file\n");
        exit(1);
    }

    fprintf(mailout, "\n");

    if (listtype > 0) {
        fprintf(mailout,
            "0This list is limited to those users who's messages make up more than %01d.%02d%%\n"
            "of total CitNet traffic.\n", listtype / 100, listtype % 100);
    } else if (listtype < 0) {
        fprintf(mailout,
            "0This list is limited to, at the most, the top %d users of CitNet,\n"
            "based upon message posting statistics.\n", -listtype);
    } else {
        fprintf(mailout,
            "0This list is the entire list of active CitNet users.\n");
    }

    fprintf(mailout, "\n"
            "3msgb0: based on the total output of the user, in bytes\n"
            "3AVb/msg0: based on average bytes per message, for the user\n"
            "3%%msgs0: based on the number user messages / total CitNet msgs\n"
            "3%%msgb0: based on the total msg bytes by that user / total CitNet msg bytes\n"
            "\n");

#ifdef XYZZY
    for (i = 0; excludenames[i].name != NULL; ++i)
        excludenames[i].hash = hash(excludenames[i].name);
#endif
}

/* -------------------------------------------------------------------- */
/*  hash()          Make an int out of their name                       */
/* -------------------------------------------------------------------- */
int hash(const char* s)
{
    int h, shift;
    int sp;

    for (h=shift=0, sp=0; s[sp]; shift=(shift+1)&7, sp++)
    {
        h ^= (toupper(s[sp])) << shift;
    }
    return h;
}

/* ------------------------------------------------------------------- */
/* config()  --  Reads a config file.                                  */
/* Lifted by GS from Audin's poster 1.6 source. Not the best config    */
/*  file routines, but they'll work for now!                           */
/* ------------------------------------------------------------------- */
void config(void)
{
    int i;
    FILE *cfg;
    char *text;

    /* we're not in a multitasker, so save stack by making them static */
    static char line[257], word1[257], lword1[257];
    static char drive[5], dir[257], dirlist[257];

    static char *KEYWORD[] = {
        "msgfile",                  /* 0 - msg.dat file */
        "outfile",                  /* 1 - output file */
        "listtype",                 /* 2 - list type */
        "exclude",                  /* 3 - user exclusion */
        "output",                   /* 4 - program output */
        NULL
    };

    drive[0] = dir[0] = dirlist[0] = 0;

    /* search these areas for the configuration file:
        current directory
        executable directory
        path
    */
    fnsplit(exename, drive, dir, NULL, NULL);
    sprintf(dirlist, ".;%s%s;%s", drive, dir, getenv("PATH"));
    cfg = pfopen("users.cfg", "rt", dirlist);
    if (cfg == NULL) {
        printf("No config file present.\n");
        exit(1);
    }

    printf("Reading config file.\n");

    while (!feof(cfg)) {
        text = fgets(line, sizeof(line), cfg);
        if(text == 0)
            break;
        if(line[0] != '#')
            continue;

        i = sscanf(line, "%s \"%255[^\"\n\r\t]", word1, lword1);

        if (i == 1)
            i = sscanf(line, "%s %255s", word1, lword1);
        if (i != 2)
            continue;

        for (i = 0; KEYWORD[i] != NULL; i++)
            if (stricmp(&word1[1], KEYWORD[i]) == 0)
                break;

        switch (i) {
            case 0:                     /* #msgfile */
                strcpy(inputmail, lword1);
                #ifndef NDEBUG
                    printf("Message file: %s\n", inputmail);
                #endif
                break;

            case 1:                     /* #outfile */
                strcpy(outputmail, lword1);
                #ifndef NDEBUG
                    printf("Output file: %s\n", outputmail);
                #endif
                break;

            case 2:                     /* #listtype */
                listtype = atoi(lword1);
                #ifndef NDEBUG
                    printf("List type: %d\n", listtype);
                #endif
                break;

            case 3:                     /* #exclude */
                {
                    xuser* newxuser;
                    lword1[31] = 0;
                    if (currxuser == NULL) {
                        /* list has no elements so far */
                        firstxuser = currxuser = (xuser *)malloc(sizeof (xuser));
                        currxuser->next = NULL;
                    }
                    else {
                        /* list has at least one element */
                        newxuser = (xuser *)malloc(sizeof (xuser));
                        newxuser->next = NULL;
                        currxuser->next = newxuser;
                        currxuser = newxuser;
                    }
                    strcpy(currxuser->name, lword1);

                    #ifndef NDEBUG
                        printf("Exclude user: %s\n", currxuser->name);
                    #endif
                }
                break;

            case 4:                     /* #user */
                output = atoi(lword1);
                #ifndef NDEBUG
                    printf("Output type: %d\n", output);
                #endif
                break;

            default:
                printf("Keyword '%s' unknown.\n", word1);
                exit(1);
                break;
        }
    }

    fclose(cfg);
}

/* from snippets 95/10
   public domain by david engel.
   slightly modified by GS.
*/
FILE *pfopen(const char* name, const char* mode, const char* dirs)
{
    char* ptr;
    char* tdirs;
    FILE* file = NULL;

    if (dirs == NULL || dirs[0] == '\0')
        return NULL;

    if ((tdirs = malloc(strlen(dirs)+1)) == NULL)
        return NULL;

    strcpy(tdirs, dirs);

    for (ptr = strtok(tdirs, ";"); file == NULL && ptr != NULL;
        ptr = strtok(NULL, ";"))
    {
        size_t len;
        char work[FILENAME_MAX];

        strcpy(work, ptr);
        len = strlen(work);
        if (len && work[len-1] != '\\')
            strcat(work, "\\");
        strcat(work, name);

        file = fopen(work, mode);
    }

    free(tdirs);

    return file;
}

/* --- End of File ---------------------------------------------------- */

